from goerr import err
from gencharts import ChartsGenerator


class Altair():
    """
    A class to handle charts with the Altair library
    """

    def __init__(self, df=None, db=None):
        """
        Initialize
        """
        self.df = df
        self.x_field = None
        self.y_field = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = None
        self.label = None
        self.engine = "bokeh"

    def altair_header_(self):
        """
        Returns html script tags for Altair
        To get the right Vega Lite library versions use this link:
        https://github.com/synw/django-chartflo/blob/master/chartflo/static/js/vegalite/
        """
        header = """
		<script src="https://rawgit.com/synw/django-chartflo/master/chartflo/static/js/vegalite/vega.js" charset="utf-8"></script>
		<script src="https://rawgit.com/synw/django-chartflo/master/chartflo/static/js/vegalite/vega-lite.js" charset="utf-8"></script>
		<script src="https://rawgit.com/synw/django-chartflo/master/chartflo/static/js/vegalite/vega-embed.js" charset="utf-8"></script>
		<style>.vega-embed canvas {max-width:100%;}</style>
		"""
        return header

    def _get_altair_chart(self, xfield, yfield, chart_type, label, opts={}, style={}, **kwargs):
        """
        Get an Altair chart object
        """
        chart = ChartsGenerator()
        # params
        opts["xfield"] = xfield
        opts["yfield"] = yfield
        opts["dataobj"] = self.df
        opts["chart_type"] = chart_type
        if "color" in style:
            opts["color"] = style["color"]
        opts["width"] = self.chart_opts["width"]
        if "height" not in self.chart_opts:
            height = 300
        else:
            height = self.chart_opts["height"]
        opts["height"] = height
        # generate
        try:
            chartobj = chart.serialize(**opts)
        except Exception as e:
            self.err(e, self._get_altair_chart,
                     "Can not get Altair chart object")
            return
        return chartobj

    def _get_altair_html(self, chart_obj, slug):
        """
        Get the html for an Altair chart
        """
        global renderer
        try:
            chart = ChartsGenerator()
            html = chart.html(slug, None, chart_obj)
            if err.exists:
                if self.errors_handling == "exceptions":
                    err.throw()
            return html
        except Exception as e:
            self.err(e, self._get_altair_html,
                     "Can not get html from the Altair rendering engine")



import pandas as pd
from goerr.colors import colors


class View():
    """
    Class to view the data
    """

    def __init__(self, df=None):
        """
        Initialize with an empty dataframe
        """
        self.df = df

    def head(self, rows=5):
        """
        Returns the main dataframe's head
        """
        if self.df is None:
            self.warning("Dataframe is empty: no head available")
            return
        return self.df.head(rows)

    def tail(self, rows=5):
        """
        Returns the main dataframe's tail
        """
        if self.df is None:
            self.warning("Dataframe is empty: no tail available")
        return self.df.tail(rows)

    def look(self, df=None, p=True):
        """
        Returns basic data info
        """
        if df is None:
            df = self.df
        if df is None:
            self.warning("Dataframe is empty: nothing to look at")
            return
        num = len(self.df.index)
        if p is True:
            print(num, "rows")
            print("Fields:", ", ".join(list(self.df)))
        else:
            return p

    def describe(self):
        """
        Return a description of the data
        """
        if self.df is None:
            self.warning("Dataframe is empty: nothing to describe")
            return
        self.look()
        return self.df.describe()

    def show(self, rows=5, dataframe=None):
        """
        Display info about a dataframe
        """
        try:
            if dataframe is not None:
                df = dataframe
            else:
                df = self.df
            if df is None:
                self.warning("Dataframe is empty: nothing to show")
                return
            num = len(df.index)
        except Exception as e:
            self.err(e, self.show, "Can not show dataframe")
            return
        f = list(df)
        num_fields = len(f)
        fds = []
        for fi in f:
            fds.append(str(fi))
        fields = ", ".join(fds)
        if self.autoprint is True:
            self.info("The dataframe has", colors.bold(num), "rows and",
                      colors.bold(num_fields), "columns:")
            print(fields)
        return df.head(rows)

    def one(self):
        """
        Shows one row of the main dataframe and the field names wiht count
        """
        try:
            return self.show(1)
        except Exception as e:
            self.err(e, self.one, "Can not display dataframe")

    def title(self, txt):
        """
        Prints a title for pipelines
        """
        num = len(txt)
        ticks = "=" * num
        print(ticks)
        print(txt)
        print(ticks)

    def subtitle(self, txt):
        """
        Prints a subtitle for pipelines
        """
        num = len(txt)
        ticks = "-" * num
        print(txt)
        print(ticks)

    def display(self, *fields):
        """
        Display some columns head
        """
        try:
            if self.df is None:
                self.warning("Dataframe is empty: nothing to display")
                return
            df2 = self.df[list(fields)]
            return df2.head()
        except Exception as e:
            self.err(e, self.display, "Can not display dataframe")

    def types_(self, col):
        """
        Display types of values in a column
        """
        df = self.df.copy()
        cols = df.columns.values
        all_types = {}
        for col in cols:
            local_types = []
            for i, val in self.df[col].iteritems():
                #print(i, val, type(val))
                t = type(val).__name__
                if t not in local_types:
                    local_types.append(t)
            all_types[col] = (local_types, i)
        df = pd.DataFrame(all_types, index=["type", "num"])
        return self.clone_(df)

    def cols(self):
        """
        Prints columns info
        """
        try:
            df = self._cols()
            return df.head(100)
        except Exception as e:
            self.err(e, self._cols, "Can not display column infos")

    def cols_(self):
        """
        Returns a DataSwim instance with columns info
        """
        try:
            df = self._cols()
            return self.clone_(df)
        except Exception as e:
            self.err(e, self._cols, "Can not display column infos")

    def _cols(self):
        """
        Returns a dataframe with columns info
        """
        try:
            df = self.df.copy()
            s = df.iloc[0]
            df = pd.DataFrame(s)
            df = df.rename(columns={0: "value"})

            # TODO : revert columns value and type order
            #cols = df.columns.tolist()
            #cols = cols[-1:] + cols[:-1]
            #df = df[cols]

            def run(row):
                t = row[0]
                return type(t).__name__

            s = df.apply(run, axis=1)
            df = df.rename(columns={0: "value"})
            df["type"] = s
            return df
        except Exception as e:
            self.err(e)

    def vals(self, col, num_col="Number"):
        """
        Print the values count of a column
        """
        df = self._vals(col, num_col)
        return df.head(0)

    def vals_(self, col, index_col=None, num_col="Number"):
        """
        Returns a DatasWim instance from values count of a column
        """
        ds2 = self.clone_(self._vals(col, num_col))
        self.quiet = True
        if index_col is None:
            index_col = col
        ds2.index_col(index_col)
        self.quiet = False
        return ds2

    def _vals(self, col, num_col):
        """
        Returns a DatasWim instance from values count of a column
        """
        if self.df is None:
            self.warning("Dataframe is empty: no values to show")
            return
        count = self.df[col].value_counts()
        df = pd.DataFrame(count)
        df = df.rename(columns={col: num_col})
        return df



import pandas as pd


class Select():
    """
    A class to retrieve data from the database
    """

    def __init__(self, db=None):
        """
        Initialize with an empty db
        """
        self.db = db

    def getall(self, table):
        """
        Get all rows values for a table
        """
        try:
            self._check_db()
        except Exception as e:
            self.err(e, self.getall, "Can not connect to database")
            return
        if table not in self.db.tables:
            self.warning("The table " + table +
                         " does not exists", self.getall)
            return
        try:
            res = self.db[table].all()
            df = pd.DataFrame(list(res))
            return df
        except Exception as e:
            self.err(e, self.getall, "Error retrieving data in table")



from goerr.colors import colors


class Info():
    """
    A class to handle information requests from the database
    """

    def __init__(self, db=None):
        """
        Initialize with an empty db
        """
        self.db = db

    def tables(self, name=None):
        """
        Print existing tables in a database
        """
        pmodels = self._tables(name)
        if pmodels is None:
            return
        if self.autoprint is True:
            num = len(pmodels)
            s = "s"
            if num == 1:
                s = ""
            msg = "Found " + colors.bold(str(num)) + " table" + s + ":\n"
            msg += "\n".join(pmodels)
            self.info(msg)

    def tables_(self, name=None):
        """
        Print existing tables in a database
        """
        pmodels = self._tables(name)
        return pmodels

    def _tables(self, name):
        """
        Print existing tables in a database
        """
        self._check_db()
        t = self.db.tables
        if name is not None:
            pmodels = [x for x in t if name in x]
        else:
            pmodels = t
        if pmodels is None:
            msg = "No tables found in the database"
            if name is not None:
                msg = "No tables found in the database for table like " + name
            self.warning(msg)
            return
        return pmodels

    def table(self, t=None, quiet=False):
        """
        Display info about a table
        """
        try:
            self._check_db()
        except Exception as e:
            self.err(e, self.table, "Can not connect to database")
            return
        if t is None:
            df = self.df
        else:
            try:
                df = self.getall(t)
            except Exception as e:
                self.err(e, self.table, "Can not get records from database")
        if self.df is None:
            self.warning("Table", t, "does not contain any record")
            return
        if self.autoprint is True and quiet is False:
            num = len(self.df)
            print(num, "rows")
            print("Fields:", ", ".join(list(df)))

    def count_rows(self, name, zero=False):
        """
        Count rows for a table
        """
        try:
            self._check_db()
        except Exception as e:
            self.err(e, self.count_rows, "Can not connect to database")
            return
        total = 0
        self.start("Counting rows in tables like", colors.bold(name))
        for t in self.tables_(name):
            num = self.db[t].count()
            total += num
            if self.autoprint is True:
                if num > 0 or zero is True:
                    msg = t + ":" + colors.bold(num)
                    print(msg)
        self.end("Found a total of", colors.bold(total), "rows")



# -*- coding: utf-8 -*-

import time
import pandas as pd
from numpy.core.numeric import nan, NaN
from goerr.colors import colors


class Clean():
    """
    Class to clean the data
    """

    def drop_nan(self, field=None, method="all"):
        """
        Drop NaN values from the main dataframe
        """
        try:
            if field is None:
                self.df = self.df.dropna(how=method)
            else:
                self.df = self.df[self.df[field].notnull()]
        except Exception as e:
            self.err(e, self.drop_nan, "Error droping nan values")

    def nan_empty(self, field):
        """
        Fill empty values with NaN values
        """
        try:
            self.df[field] = self.df[field].replace('', nan)
        except Exception as e:
            self.err(e, self.nan_empty_, "Can not fill empty values with nan")

    def zero_nan(self, *fields):
        """
        Converts zero values to nan values in selected columns
        """
        try:
            self.df = self._zero_nan(*fields)
        except Exception as e:
            self.err(e, self.nan_empty_, "Can not fill zero values with nan")

    def zero_nan_(self, *fields):
        """
        Returns a DataSwim instance with zero values to nan values in selected columns
        """
        df = self.clone_(self._zero_nan(*fields))
        return self.clone_(df)

    def _zero_nan(self, *fields):
        """
        Converts zero values to nan values in selected columns
        """
        df = self.df.copy()
        try:
            for field in fields:
                df = df.replace({field: {nan: 0}})
            if self.autoprint is True:
                s = "s"
                if len(fields) == 1:
                    s = ""
                self.ok("Replaced 0 values by nan in column" + s, field)
            return df
        except Exception as e:
            self.err(e)

    def fill_nan(self, val, *fields):
        """
        Fill NaN values with new values
        """
        try:
            self.df = self._fill_nan(val, *fields)
        except Exception as e:
            self.err(e, self.fill_nan, "Can not fill nan values")

    def fill_nan_(self, val, *fields):
        """
        Returns a DataSwim instance with NaN values filled
        """
        fields = list(fields)
        if len(fields) == 0:
            fields = self.df.index.values
        try:
            df = self._fill_nan(val, *fields)
            return self.clone_(df=df)
        except Exception as e:
            self.err(e, self.fill_nan_, "Can not fill nan values")

    def _fill_nan(self, val, *fields):
        """
        Fill NaN values with new values
        """
        fields = list(fields)
        df = self.df.copy()
        try:
            for el in fields:
                try:
                    df[el] = df[[el]].fillna(val)
                except KeyError:
                    self.warning("Can not find column ", el)
                    return
        except Exception as e:
            self.err(e, self._fill_nan, "Can not fill nan values")
        if self.autoprint is True:
            self.ok("Filled nan values in columns", *fields)
        return df

    def replace(self, col, searchval, replaceval):
        """
        Replace a value in a column in the main dataframe
        """
        try:
            self.df = self._replace(col, searchval, replaceval)
        except Exception as e:
            self.err(e, self.replace, "Can not replace value in column")

    def replace_(self, col, searchval, replaceval):
        """
        Returns a Dataswim instance with replaced values in a column
        """
        try:
            return self._replace(col, searchval, replaceval)
        except Exception as e:
            self.err(e, self.replace_, "Can not replace value in column")

    def _replace(self, col, searchval, replaceval):
        """
        Replace a value in a colum
        """
        df = self.df.copy()
        try:
            df[col] = df[col].replace(searchval, replaceval)
            return df
        except Exception as e:
            self.err(e, self._replace, "Can not replace value in column")

    def fill_nulls(self, field):
        """
        Fill all null values with NaN values
        """
        n = [None, ""]
        try:
            self.df[field] = self.df[field].replace(n, nan)
        except Exception as e:
            self.err(e)

    def to_int(self, field):
        """
        Convert some column values to integers
        """
        """
        for field in fields:
            self.drop_nan(field)
            self.df[field] = self.df[field].dropna().astype(int)

        """
        ds2 = self.clone_()

        def convert(val):
            try:
                return int(val)
            except Exception as e:
                print(val, str(type(val)))
                raise(e)
                return val
        try:
            ds2.df[field] = ds2.df[field].apply(convert)
            self.df = ds2.df
        except Exception as e:
            self.err(e, self.to_int, "Can not convert column values to integer")
            return
        if self.autoprint is True:
            self.ok("Converted column values to integers")

    def to_float(self, *cols):
        """
        Convert colums values to float
        """
        try:
            df = self.to_type("float64", *cols)
            self.df = df
        except Exception as e:
            self.err(e, self.to_type, "Can not convert column values to float")
            return
        if self.autoprint is True:
            self.ok("Converted column values to float")

    def to_type(self, dtype, *cols):
        """
        Convert colums values to a given type
        """
        df = self.df.copy()
        allcols = df.columns.values
        try:
            for col in cols:
                if col not in allcols:
                    self.err(self.to_type, "Column " + col + " not found")
                    return
                df[col] = df[col].astype(dtype)
            return df
        except Exception as e:
            self.err(e)

    def clean_ts(self, date_col, numeric_col=None, index=True, to_int=False, index_col=True):
        """
        Cleans and format a timeseries dataframe
        """
        try:
            self.date(date_col)
            if index is True:
                self.dateindex(date_col)
            if numeric_col is not None:
                self.fill_nan(numeric_col)
                if to_int is True:
                    self.to_int(numeric_col)
            if index_col is True:
                self.index_col()
            self.index_col(date_col)
        except Exception as e:
            self.err(e)

    def date(self, *fields, precision="S"):
        """
        Convert column values to properly formated datetime
        """
        def convert(row):
            try:
                t1 = row.timetuple()
                int(time.mktime(t1))
            except:
                return nan
            encoded = '%Y-%m-%d %H:%M:%S'
            if precision == "Min":
                encoded = '%Y-%m-%d %H:%M'
            elif precision == "H":
                encoded = '%Y-%m-%d %H'
            elif precision == "D":
                encoded = '%Y-%m-%d'
            elif precision == "M":
                encoded = '%Y-%m'
            elif precision == "Y":
                encoded = '%Y'
            return row.strftime(encoded)

        try:
            for f in fields:
                try:
                    df2 = self.df.copy()
                    self.df[f] = pd.to_datetime(df2[f]).apply(convert)
                except ValueError:
                    pass
        except KeyError:
            self.warning("Can not find colums " + " ".join(fields))
            return
        except Exception as e:
            self.err(e, self.date, "Can not process date field")

    def dateindex(self, datafield, indexfield="date_index", df=None):
        """
        Set a datetime index from a column
        """
        try:
            self.df = self._dateindex(datafield, indexfield, df)
        except Exception as e:
            self.err(e, self.dateindex, "Can not index data")

    def dateindex_(self, datafield, indexfield="date_index", df=None):
        """
        Returns a DataSwim instance index from a column
        """
        try:
            df = self._dateindex(datafield, indexfield, df)
        except Exception as e:
            self.err(e, self.dateindex_, "Can not index data")
            return
        return self.clone_(df=df)

    def _dateindex(self, datafield, indexfield, df=None):
        """
        Returns a datetime index from a column
        """
        if df is None:
            df = self.df.copy()
        try:
            index = pd.DatetimeIndex(df[datafield])
            df = df.set_index(index)
        except KeyError as e:
            self.err(e, self._dateindex, "Can not find column " +
                     colors.bold(datafield) + " in data")
            return
        except Exception as e:
            msg = "Can not reindex with datafield " + \
                colors.bold(datafield) + " and indexfield " + \
                colors.bold(indexfield)
            self.err(e, self._dateindex, msg)
            return
        if self.autoprint is True:
            self.ok("Added a datetime index from column", datafield)
        return df

    def index(self, indexcol):
        """
        Set an index to the main dataframe
        """
        try:
            self.df = self._index(indexcol)
        except KeyError as e:
            self.err(e, self.index, "Can not create index")

    def index_(self, indexcol):
        """
        Returns a Dataswim instance with an index
        """
        try:
            return self._index(indexcol)
        except KeyError as e:
            self.err(e, self.index_, "Can not create index")

    def _index(self, indexcol):
        """
        Set an index to the main dataframe
        """
        df = self.df.copy()
        try:
            df = df.set_index(df[indexcol])
        except KeyError as e:
            self.err(e, self._index, "Can not find column " +
                     colors.bold(indexcol) + " in data")
            return
        except Exception as e:
            msg = "Can not index with columns " + \
                colors.bold(indexcol) + " and indexfield "
            self.err(e, self._index, msg)
            return
        if self.autoprint is True:
            self.ok("Added an index from column", indexcol)
        return df

    def strip(self, col):
        """
        Remove white space in a column's values
        """
        def remove_ws(row):
            val = str(row[col])
            if " " in val:
                row[col] = val.replace(" ", "")
            return row

        try:
            self.apply(remove_ws)
        except Exception as e:
            self.err(e, self.strip, "Can not remove white space in column")
            return
        if self.autoprint is True:
            self.ok("White space removed in column values")

    def strip_cols(self):
        """
        Remove white space in columns names
        """
        try:
            cols = {}
            skipped = []
            for col in self.df.columns.values:
                try:
                    cols[col] = col.strip()
                except:
                    skipped.append(str(col))
            self.df = self.df.rename(columns=cols)
        except Exception as e:
            self.err(e, self.strip_cols, "Can not strip white space in columns")
            return
        if self.autoprint is True:
            self.ok("White space removed in columns names")
            if len(skipped) > 0:
                self.info("Skipped columns", ','.join(
                    skipped), "while removing white space")

    def format_date_(self, date):
        """
        Formats a date
        """
        return date.strftime('%Y-%m-%d %H:%M:%S')

    def transform_(self, dateindex=None, index_col=None, fill_col=None, num_col=None, df=None):
        """
        Returns a DataSwim instance transformed according to the given parameters
        """
        if df is None:
            if self.df is None:
                self.err(
                    self._transform, "No dataframe: please provide one in parameters or set it")
                return
            df = self.df.copy()
        ds2 = self.clone_(df)
        if dateindex is None and index_col is None and fill_col is None and num_col is None:
            return ds2
        try:
            if dateindex is not None:
                ds2 = ds2.dateindex_(dateindex, df=df)
            if fill_col is not None:
                ds2 = ds2.fill_nan_(0, fill_col)
            if index_col is not None:
                ds2 = ds2.index_col_(index_col)
            if num_col is not None:
                ds2.add(num_col, 1)
            return ds2
        except Exception as e:
            self.err(e, self.transform_, "Can not transform data")



palette = ["#15527F", "#FF9900", "#FFCC00", "#0099CC", "#999900", "#663366",
           "#FF0000", "#33CC99", "#006600", "#663300", "#99CC00", "#FF0033"]
color_num = 0


class Colors():
    """
    Class to handle colors for charts
    """

    def color_(self, i=None):
        """
        Get a color from the palette
        """
        global palette, color_num
        if i is not None:
            color_num = i
        if color_num == len(palette) - 1:
            color_num = 0
        res = palette[color_num]
        color_num += 1
        return res





class Insert():
    """
    A class to handle data ingestion by the database
    """

    def __init__(self, db=None):
        """
        Initialize with an empty db
        """
        self.db = db

    def insert(self, table, records, create_cols=True):
        """
        Insert one or many records in the database from a dictionary or a list of dictionaries
        """
        self._check_db()
        try:
            table = self.db[table]
        except Exception as e:
            self.err(e, self.update_db, "Can not find table " + table)
        t = type(records)
        if t == dict:
            func = table.insert
        elif t == list:
            func = table.insert_many
        else:
            msg = "Rows datatype " + \
                str(t) + " not valid: use a list or a dictionary"
            self.err(self.insert, msg)
        if create_cols is True:
            try:
                func(records, ensure=True)
            except Exception as e:
                self.err(e, self.insert,
                         "Can not insert create columns and insert data")
            return
        else:
            try:
                func(records)
            except Exception as e:
                self.err(e, self.insert,
                         "Can not insert create columns and insert data")
            return
        if self.autoprint is True:
            self.ok("Rows inserted in the database")

    def update_db(self, table, keys=['id'], db_url=None):
        """
        Update records in a database table from the main dataframe
        """
        try:
            table = self.db[table]
        except Exception as e:
            self.err(e, self.update_db, "Can not find table " + table)
        if self.db is None and db_url is None:
            msg = "Please connect a database before or provide a database url"
            self.err(self.to_db, msg)
            return
        recs = self.to_records_()
        for rec in recs:
            table.insert_ignore(rec, keys, ensure=True)
        if self.autoprint is True:
            self.ok("Data updated in table", table)

    def to_db(self, table, db_url=None):
        """
        Save the main dataframe to the database
        """
        if self.db is None and db_url is None:
            msg = "Please connect a database before or provide a database url"
            self.err(self.to_db, msg)
            return
        if db_url is not None:
            self.connect(db_url)
        recs = self.to_records_()
        self.insert(table, recs)
        if self.autoprint is True:
            self.ok("Data inserted in table", table)



import datetime
import dateutil.relativedelta
from IPython.display import display, HTML
from goerr.colors import colors


class Messages():
    """
    A class to handle output
    """

    def ok(self, *msg):
        """
        Prints a message with an ok prefix
        """
        label = colors.green("OK")
        self._msg(label, *msg)

    def info(self, *msg):
        """
        Prints a message with an info prefix
        """
        label = colors.blue("INFO")
        endmsg = self.msg_(label, *msg)
        print(endmsg)

    def warning(self, *msg):
        """
        Prints a warning
        """
        label = colors.yellow("WARNING")
        self._msg(label, *msg)

    def debug(self, *msg):
        """
        Prints a warning
        """
        label = colors.yellow("DEBUG")
        self._msg(label, *msg)

    def progress(self, *msg):
        """
        Prints a progress message
        """
        label = colors.purple("Progress")
        self._msg(label, *msg)

    def start(self, *msg):
        """
        Prints an start message
        """
        self.start_time = datetime.datetime.now()
        label = colors.purple("START")
        self._msg(label, *msg)

    def end(self, *msg):
        """
        Prints an end message with elapsed time
        """
        if self.start_time is None:
            self.err(
                self.end, "No start time set: please use start() before using this function")
        endtime = datetime.datetime.now()
        rd = dateutil.relativedelta.relativedelta(endtime, self.start_time)
        endmsg = self._endmsg(rd)
        label = colors.purple("END")
        msg += ("in " + endmsg,)
        self._msg(label, *msg)
        self.start_time = None

    def html(self, label, *msg):
        """
        Prints html in notebook
        """
        lbl = "[" + label + "] "
        txt = lbl + " " + " ".join(list(msg))
        if self.notebook is True:
            html = HTML(txt)
            display(lbl + html)
        else:
            print(lbl + txt)

    def msg_(self, label, *msg):
        """
        Returns a message with a label
        """
        txt = self._unpack_msg(*msg)
        return "[" + label + "] " + txt

    def _msg(self, label, *msg):
        """
        Prints a message with a label
        """
        if self.quiet is False:
            txt = self._unpack_msg(*msg)
            print("[" + label + "] " + txt)

    def _unpack_msg(self, *msg):
        """
        Convert all message elements to string
        """
        l = []
        for m in msg:
            l.append(str(m))
        return " ".join(l)

    def _endmsg(self, rd):
        """
        Returns an end message with elapsed time
        """
        msg = ""
        s = ""
        if rd.hours > 0:
            if rd.hours > 1:
                s = "s"
            msg += colors.bold(str(rd.hours)) + " hour" + s + " "
        s = ""
        if rd.minutes > 0:
            if rd.minutes > 1:
                s = "s"
            msg += colors.bold(str(rd.minutes)) + " minute" + s + " "
        # if rd.seconds > 0:
        #    msg+=str(rd.seconds)
        # else:
        #    msg+="0."
        milliseconds = int(rd.microseconds / 1000)
        if milliseconds > 0:
            msg += colors.bold(str(rd.seconds) + "." + str(milliseconds))
        msg += " seconds"
        return msg



from .bokeh import Bokeh
from .altair import Altair
from .chartjs import Chartjs
from .seaborn import Seaborn
from .colors import Colors


class Plot(Bokeh, Altair, Chartjs, Seaborn, Colors):
    """
    Class to handle charts
    """

    def __init__(self, df=None):
        """
        Initialize
        """
        self.df = df
        self.x = None
        self.err = None
        self.y = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = {}
        self.label = None
        self.engine = "bokeh"

    def chart(self, x=None, y=None, chart_type=None, opts=None,
              style=None, label=None, options={}, **kwargs):
        """
        Get a chart
        """
        try:
            self.chart_obj = self._chart(x, y, chart_type, opts, style, label,
                                         options=options, **kwargs)
        except Exception as e:
            self.err(e, self.chart, "Can not create chart")

    def chart_(self, x=None, y=None, chart_type=None, opts=None, style=None,
               label=None, options={}, **kwargs):
        """
        Get a chart
        """
        try:
            return self._chart(x, y, chart_type, opts, style, label,
                               options=options, **kwargs)
        except Exception as e:
            self.err(e, self.chart, "Can not create chart")

    def _chart(self, x, y, chart_type, opts, style, label, options, **kwargs):
        """
        Initialize chart options
        """
        if opts is not None:
            self.chart_opts = opts
        if style is not None:
            self.chart_style = style
        if label is not None:
            self.label = label
        self.x = x
        self.y = y
        if chart_type is None:
            return
        try:
            chart_obj = self._get_chart(chart_type, x, y, style=style,
                                        opts=opts, label=label,
                                        options=options, **kwargs)
            return chart_obj
        except Exception as e:
            self.err(e)

    def bar_(self, label=None, style=None, opts=None, options={}):
        """
        Get a bar chart
        """
        try:
            return self._get_chart("bar", style=style, opts=opts, label=label,
                                   options=options)
        except Exception as e:
            self.err(e, self.chart, "Can draw bar chart")

    def line_(self, label=None, style=None, opts=None, options={}):
        """
        Get a line chart
        """
        try:
            return self._get_chart("line", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.line_, "Can draw line chart")

    def area_(self, label=None, style=None, opts=None, options={}):
        """
        Get an area chart
        """
        try:
            return self._get_chart("area", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.area_, "Can draw area chart")

    def hist_(self, label=None, style=None, opts=None, options={}):
        """
        Get an historiogram chart
        """
        try:
            return self._get_chart("hist", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.hist_, "Can draw historiogram")

    def errorbar_(self, label=None, style=None, opts=None, options={}):
        """
        Get a point chart
        """
        try:
            return self._get_chart("errorBar", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.errorbar_, "Can draw errorbar chart")

    def point_(self, label=None, style=None, opts=None, options={}):
        """
        Get a point chart
        """
        try:
            return self._get_chart("point", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.point_, "Can draw point chart")

    def circle_(self, label=None, style=None, opts=None, options={}):
        """
        Get a circle chart
        """
        try:
            return self._get_chart("circle", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.point_, "Can draw circle chart")

    def square_(self, label=None, style=None, opts=None, options={}):
        """
        Get a square chart
        """
        try:
            return self._get_chart("square", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.point_, "Can draw square chart")

    def tick_(self, label=None, style=None, opts=None, options={}):
        """
        Get a tick chart
        """
        try:
            return self._get_chart("tick", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.point_, "Can draw tick chart")

    def rule_(self, label=None, style=None, opts=None, options={}):
        """
        Get a rule chart
        """
        try:
            return self._get_chart("rule", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.point_, "Can draw rule chart")

    def heatmap_(self, label=None, style=None, opts=None, options={}):
        """
        Get a heatmap chart
        """
        try:
            return self._get_chart("heatmap", style=style, opts=opts,
                                   label=label, options=options)
        except Exception as e:
            self.err(e, self.heatmap_, "Can draw heatmap")

    def line_point_(self, label=None, style=None, opts=None, options={},
                    colors={"line": "yellow", "point": "navy"}):
        """
        Get a line and point chart
        """
        try:
            if style is None:
                style = self.chart_style
            style["color"] = colors["line"]
            c = self._get_chart("line", style=style, opts=opts,
                                label=label, options=options)
            style["color"] = colors["point"]
            c2 = self._get_chart("point", style=style, opts=opts,
                                 label=label, options=options)
            return c * c2
        except Exception as e:
            self.err(e, self.line_point_, "Can draw line_point chart")

    def opts(self, dictobj):
        """
        Add or update an option value to defaults
        """
        for k in dictobj:
            self.chart_opts[k] = dictobj[k]

    def style(self, dictobj):
        """
        Add or update a style value to defaults
        """
        for k in dictobj:
            self.chart_style[k] = dictobj[k]

    def color(self, val):
        """
        Change the chart's color
        """
        self.style(dict(color=val))

    def width(self, val):
        """
        Change the chart's width
        """
        self.opts(dict(width=val))

    def height(self, val):
        """
        Change the chart's height
        """
        self.opts(dict(height=val))

    def size(self, val):
        """
        Change the chart's point size
        """
        self.style(dict(size=val))

    def reset_opts(self):
        """
        Reset the chart options
        """
        self.chart_opts = {}

    def reset_style(self):
        """
        Reset the chart options
        """
        self.chart_style = {}

    def _get_chart(self, chart_type, x=None, y=None, style=None, opts=None,
                   label=None, options={}, **kwargs):
        """
        Get a full chart object
        """
        if x is None:
            if self.x is None:
                self.err(
                    self._get_chart,
                    "X field is not set: please specify a parameter")
                return
            x = self.x
        if y is None:
            if self.y is None:
                self.err(
                    self._get_chart,
                    "Y field is not set: please specify a parameter")
                return
            y = self.y
        if opts is None:
            opts = self.chart_opts
        if style is None:
            style = self.chart_style
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if self.engine == "bokeh":
            func = self._get_bokeh_chart
        elif self.engine == "altair":
            func = self._get_altair_chart
        elif self.engine == "chartjs":
            func = self._get_chartjs_chart
        elif self.engine == "seaborn":
            func = self._get_seaborn_chart
        else:
            self.err(self._get_chart, "Engine " + self.engine + " unknown")
            return
        try:
            chart = func(
                x, y, chart_type, label, opts, style,
                options=options, **kwargs)
            return chart
        except Exception as e:
            self.err(e)

    def _check_defaults(self, x_only=True):
        """
        Checks if charts defaults are set
        """
        if self.x is None:
            self.err(
                self.linear_, "X field is not set: please specify a parameter")
            return
        if x_only is True:
            return
        if self.y is None:
            self.err(
                self.linear_, "Y field is not set: please specify a parameter")
            return



import holoviews as hv
from bokeh.embed import components
from holoviews.core.data.interface import DataError


bokeh_renderer = hv.renderer('bokeh')


class Bokeh():
    """
    A class to handle charts with the Bokeh library
    """

    def __init__(self, df=None, db=None):
        """
        Initialize
        """
        self.df = df
        self.x_field = None
        self.y_field = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = None
        self.label = None
        self.engine = "bokeh"

    def ndlayout_(self, dataset, kdims, cols=3):
        """
        Create a Holoview NdLayout from a dictionnary of chart objects
        """
        try:
            return hv.NdLayout(dataset, kdims=kdims).cols(cols)
        except Exception as e:
            self.err(e, self.layout_, "Can not create layout")

    def layout_(self, chart_objs, cols=3):
        """
        Returns a Holoview Layout from chart objects
        """
        try:
            return hv.Layout(chart_objs).cols(cols)
        except Exception as e:
            self.err(e, self.layout_, "Can not build layout")

    def bokeh_header_(self):
        """
        Returns html script tags for Bokeh
        """
        header = """
        <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/bokeh/0.12.10/bokeh.min.js"></script>         
        <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/bokeh/0.12.10/bokeh.min.js.map"></script>
        <script type="text/javascript">
            Bokeh.set_log_level("info");
        </script>
        """
        return header

    def hline_(self, col, label=None):
        """
        Returns an horizontal line from a column mean value
        """
        c = hv.HLine(self.df[col].mean())
        return c

    def _get_bokeh_chart(self, x_field, y_field, chart_type, label, opts, style, options={}, **kwargs):
        """
        Get a Bokeh chart object
        """
        if type(x_field) == list:
            kdims = x_field
        else:
            kdims = [x_field]
        if type(y_field) == list:
            vdims = y_field
        else:
            vdims = [y_field]
        args = kwargs
        args["data"] = self.df
        args["kdims"] = kdims
        args["vdims"] = vdims
        if label is not None:
            args["label"] = label
        else:
            if self.label is not None:
                args["label"] = self.label
        chart = None
        try:
            if chart_type == "line":
                chart = hv.Curve(**args)
            elif chart_type == "point":
                chart = hv.Scatter(**args)
            elif chart_type == "area":
                chart = hv.Area(**args)
            elif chart_type == "bar":
                chart = hv.Bars(**args)
            elif chart_type == "hist":
                chart = hv.Histogram(**args)
            elif chart_type == "errorBar":
                chart = hv.ErrorBars(**args)
            endchart = chart(plot=opts, style=style)
            return endchart
            if chart is None:
                self.err("Chart type " + chart_type +
                         " unknown", self._get_bokeh_chart)
                return
        except DataError as e:
            msg = "Column not found in " + x_field + " and " + y_field
            self.err(e, self._get_bokeh_chart, msg)
        except Exception as e:
            self.err(e)

    def _get_bokeh_html(self, chart_obj):
        """
        Get the html for a Bokeh chart
        """
        global bokeh_renderer
        try:
            renderer = bokeh_renderer
            p = renderer.get_plot(chart_obj).state
            script, div = components(p)
            return script + "\n" + div

        except Exception as e:
            self.err(e, self._get_bokeh_html,
                     "Can not get html from the Bokeh rendering engine")



# -*- coding: utf-8 -*-

from numpy import where


class Count():
    """
    Class to count data
    """

    def __init__(self, df=None):
        """
        Initialize with an empty dataframe
        """
        self.df = df

    def count_nulls(self, field):
        """
        Count the number of null values in a rows
        """
        try:
            n = self.df[field].isnull().sum()
        except KeyError:
            self.warning("Can not find column", field)
            return
        except Exception as e:
            self.err(e, self.count_nulls, "Can not count nulls")
            return
        if self.autoprint is True:
            self.ok("Found", n, "nulls in column", field)

    def count(self):
        """
        Counts the number of rows of the main dataframe
        """
        try:
            num = self._count()
        except Exception as e:
            self.err(e, self.count, "Can not count data")
            return
        if self.autoprint is True:
            self.ok("Found", num, "rows in the dataframe")

    def count_(self):
        """
        Returns the number of rows of the main dataframe
        """
        try:
            num = self._count()
        except Exception as e:
            self.err(e, self.count_, "Can not count data")
            return
        return num

    def _count(self):
        """
        Count the number of rows of the main dataframe
        """
        try:
            num = len(self.df.index)
            return num
        except Exception as e:
            self.err(e)
            return

    def count_empty(self, field):
        """
        List of empty row indices
        """
        try:
            df2 = self.keep_(field).df
            vals = where(df2.applymap(lambda x: x == ''))
            num = len(vals[0])
            return num
        except Exception as e:
            self.err(e, self.count_empty, "Can not count empty values")
            return
        if self.autoprint is True:
            self.ok("Found", num, "empty rows in the dataframe")

    def count_zero(self, field):
        """
        List of row with 0 values
        """
        try:
            df2 = self.keep_(field).df
            vals = where(df2.applymap(lambda x: x == 0))
            num = len(vals[0])
            return num
        except Exception as e:
            self.err(e, self.count_zero, "Can not count zero values")
        if self.autoprint is True:
            self.ok("Found", num, "zero values rows in column", field)

    def count_unique(self, field):
        """
        Return the number of unique values in a column     
        """
        try:
            num = self.df[field].nunique()
            return num
        except Exception as e:
            self.err(e, self.count_unique, "Can not count unique values")
        if self.autoprint is True:
            self.ok("Found", num, "unique values rows in column", field)
